@@ -1,3 +1,4 @@
+################################################################################
 """
 pycev: A compressed event record reader for SEL CEV files.
 
@@ -6,9 +7,13 @@ Author(s):
 
 Homepage: https://github.com/engineerjoe440/pycev
 
-SEL Protocol Application Guide: https://selinc.com/api/download/5026/
-SEL Compressed ASCII (CASCII) Application Guide: https://selinc.com/api/download/5021/
---------------------------------------------------------------------------------------
+SEL Protocol Application Guide:
+    https://selinc.com/api/download/5026/
+
+SEL Compressed ASCII (CASCII) Application Guide:
+    https://selinc.com/api/download/5021/
+
+───────────────────────────────────────────────────────────────────────────────
 
 MIT License
 
@@ -32,129 +37,52 @@ LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE.
 """
+################################################################################
 
 # Standard Imports
 import os
 import re
-import time
 import inspect
+from typing import Union
 import warnings
 import datetime as dt
 from collections import namedtuple
 
-# Local Imports
-
 # Describe Package for External Interpretation
 _name_ = "pycev"
-_version_ = "0.0"
+_version_ = "0.1.0"
 __version__ = _version_  # Alias the Version String
 
 
+# Define Global Strings
 EVENT_SETTINGS_SEP = '"SETTINGS","02E1"'
+ANALOG_SAMPLES = "SAM/CYC_A"
+DIGITAL_SAMPLES = "SAM/CYC_D"
+TRIGGER_KEY_CHAR = ">"
+FREQUENCY_KEY = "FREQ"
 
 
-# Define Function to Evaluate Checksum
-def _eval_checksum( data, constrain=True ):
-    """
-    *Evaluate the expected checksum for specific data row*
+# Custom Exceptions and Warnings
+class MalformedChecksumFailure(UserWarning):
+    """Record data appears to be malformed, and fails checksum validation."""
 
-    This function accepts a string, and calculates the checksum
-    of the bytes provided.
+class UnexpectedFileExtension(UserWarning):
+    """File does not appear to use expected "CEV" extension."""
 
-    Parameters
-    ----------
-    data:       [str, bytes]
-                The bytestring which should be evaluated for the
-                checksum.
-    constrain:  bool, optional
-                Control to specify whether the value should be
-                constrained to an 8-bit representation, defaults
-                to False.
-    
-    Returns
-    -------
-    checksum:   int
-                The fully evaluated checksum.
-    """
-    # Evaluate the sum
-    if isinstance(data, str):
-        checksum = sum(map(ord, data))
-    else:
-        checksum = sum(data)
-    # Cap the Value if Needed
-    if constrain:
-        checksum = checksum & 0xffff # Bit-wise AND with 16-bit maximum
-    return checksum
+class MalformedHeadingDataMismatch(Exception):
+    """CEV is Malformed - Heading and Data Section Lengths do not Match."""
 
-# Define Function to Cast Hex Byte to "bool-like" List
-def hex_byte_to_bits(hex_byte):
-    """
-    *Interpret a single hex character as bits*
-    
-    This function will accept a hex byte (two characters, 0-F)
-    and cast it to a list of eight bits. Credit to StackOverflow
-    user: "Beebs" for concise function implementation.
-    
-    Parameters
-    ----------
-    hex_byte:   str
-                String of two characters representing the byte
-                in hex.
-    
-    Returns
-    -------
-    bits:       list of int
-                List of eight bits (0/1) representing each bit
-                position in the input byte.
-    
-    See Also
-    --------
-    hex_bits_from_str   : Determine the bit-structure of multiple
-                          hex bytes in a string.
-    """
-    binary_byte = bin(int(hex_byte, base=16))
-    # Use zfill to pad the string with zeroes as we want all 8 digits of the byte.
-    bits_string = binary_byte[2:].zfill(8)
-    return [int(bit) for bit in bits_string]
+class MalformedNoFIDFound(Exception):
+    """CEV is Malformed - FID String Could not be Located."""
+
+class MalformedNoSampleNumberFound(Exception):
+    """CEV is Malformed - Number of Samples Could not be Located."""
 
-# Define Function to Evaluate List of Bits from Raw Hex String
-def hex_bits_from_str(hex_string):
-    """
-    *Interpret string of hex characters as a list of bits*
-    
-    This function will accept the hex string presented in CEV
-    files and construct a list of bits where each bit corresponds
-    directly to one of the relay wordbits.
-    
-    Parameters
-    ----------
-    hex_string: str
-                String of hex characters to be interpreted.
-    
-    Returns
-    -------
-    bits:       list of int
-                List of bits (int; 0/1) as interpreted from
-                the hex string input.
-        
-    See Also
-    --------
-    hex_byte_to_bits    : Determine the bit-structure of a single
-                          hex byte
-    """
-    bits = []
-    # Iteratively Process the Hex String
-    while len(hex_string) > 0:
-        hex_byte    = hex_string[:2]
-        hex_string  = hex_string[2:]
-        # Identify the Bits
-        bits.extend( hex_byte_to_bits(hex_byte) )
-    return bits
 
 # Define Function to Interpret Row-Wise Checksum for Validity
-def row_wise_checksum(row_data):
+def row_wise_checksum(row_data, constrain=True):
     """
-    *Identify the data and validate it with included checksum*
+    Identify the data and validate it with included checksum.
 
     Events should contain rows of data, where each row appears
     as:
@@ -171,7 +99,7 @@ def row_wise_checksum(row_data):
     row_data:   str
                 The string containing all row data in form
                 described above.
-    
+
     Returns
     -------
     row:        namedtuple
@@ -182,28 +110,38 @@ def row_wise_checksum(row_data):
     # Define the namedtuple Structure used for Return
     Row = namedtuple('Row', ['data', 'validity'])
     # Clean Input Data
-    row_data = row_data.replace('\r','')
-    row_data = row_data.replace('\n','')
+    row_data = row_data.replace('\r', '')
+    row_data = row_data.replace('\n', '')
     # Successfully Extracted Data
-    row_contents = row_data[:-6] # Remove Checksum Characters
-    checksum_str = row_data[-6:].replace('"','') # Keep Checksum Only
+    row_contents = row_data[:-6]  # Remove Checksum Characters
+    checksum_str = row_data[-6:].replace('"', '')  # Keep Checksum Only
     # Evaluate Checksum as INT
-    checksum_int = int.from_bytes(  bytes.fromhex(checksum_str),
-                                    byteorder='big',
-                                    signed=True )
-    # Calculate Checksum
-    checksum = _eval_checksum(data=row_contents)
+    checksum_int = int.from_bytes(
+        bytes.fromhex(checksum_str),
+        byteorder='big',
+        signed=False
+    )
+    # Evaluate the Checksum
+    if isinstance(row_contents, str):
+        checksum = sum(map(ord, row_contents))
+    else:
+        checksum = sum(row_contents)
+    # Cap the Value if Needed
+    if constrain:
+        checksum = checksum & 0xffff  # Bit-wise AND with 16-bit maximum
     # Remove Trailing Comma if Present
     if row_contents.endswith(','):
-        row_contents = row_contents[:-1] # Trim Comma
+        row_contents = row_contents[:-1]  # Trim Comma
     # Pack Structure
-    row = Row(data=row_contents, validity=(checksum==checksum_int))
+    valid = checksum == checksum_int
+    row = Row(data=row_contents, validity=valid)
     return row
 
+
 # Define Function to Split Event Data from Relay Settings
 def split_event_and_relay_data(data):
     """
-    *Split the Event Data from Relay Settings*
+    Split the Event Data from Relay Settings.
 
     This function accepts the full text from a CEV file, and separates
     the event-record data from the relay's configuration settings.
@@ -214,7 +152,7 @@ def split_event_and_relay_data(data):
     ----------
     data:       str
                 The string containing all data read from the CEV file.
-    
+
     Returns
     -------
     event:      namedtuple
@@ -237,8 +175,8 @@ def split_event_and_relay_data(data):
 # Define the Primary Class
 class Cev():
     """
-    *SEL CEV File Reader*
-    
+    SEL CEV File Reader.
+
     This class serves to provide functionality to read SEL (Schweitzer
     Engineering Laboratories) Compressed EVent records (CEV files) and
     grant pragmatic access to the key components of these records. The
@@ -248,11 +186,11 @@ class Cev():
     conveniently interpret information. This all is in an effort to
     support data scientists and engineers to make empowered, intelligent
     decisions about the power systems which they are responsible for.
-    
+
     Within this module, this class (`Cev`) is also aliased as CEV for
     programming convenience, and to allow users to access the functionality
     in a format that is consistent with their work.
-    
+
     Parameters
     ----------
     file:       str, optional
@@ -267,17 +205,17 @@ class Cev():
     **kwargs:   dict, optional
                 Additional optional parameters which may be specified
                 during class initialization. Common arguments include:
-                
+
                     - encoding:     str, file encoding such as 'utf-8'
                     - decode_opt:   str, file decoding option such as
                                     'strict', 'ignore', 'replace', or
                                     'backslashreplace' as defined by
                                     standard str.decode method.
                     - ignore_warnings: bool, control to ignore warnings
-                
+
                 Class initialization may include one or more kwargs,
                 but none are required.
-    
+
     Attributes
     ----------
     analog_channels:        list of list of float
@@ -345,10 +283,20 @@ class Cev():
     trigger_time:           datetime
                             Date-time structure indicating when the event
                             was "triggered" by protection logic in the relay.
+
+    Examples
+    --------
+    >>> from pycev import CEV
+    >>> # Load a file and parse, directly.
+    >>> record = CEV(file="./event-report.cev")
+    >>> print("Trigger time = {}s".format(record.trigger_time))
     """
-    
-    def __init__(self, file=None, data=None, **kwargs):
+
+    def __init__(self, file: str = None, data: Union[str, bytes] = None,
+                 **kwargs):
         """
+        Prepare CEV Reader.
+
         Class initialization with optional data input methods for file-path
         and raw data as either a string or bytes.
         """
@@ -365,13 +313,13 @@ class Cev():
             self._decode_opt = kwargs['decode_option']
         else:
             self._decode_opt = 'strict'
-        
+
         # Handle Warnings Override
         if "ignore_warnings" in kwargs:
             self.ignore_warnings = kwargs["ignore_warnings"]
         else:
             self.ignore_warnings = False
-        
+
         # Prepare Defaults
         self.data = ''
         self.record = ''
@@ -379,109 +327,132 @@ class Cev():
         self.record_lines = []
         self.fid = ''
         self.raw_fid = ''
-        self.trigger_time = dt.datetime(1970, 1, 1) # Default to Epoch
+        self.time = []
+        self.trigger_time = dt.datetime(1970, 1, 1)  # Default to Epoch
         self.channels_count = 0
         self.analog_channels = []
         self.analog_channel_ids = []
         self.status_channels = []
         self.status_channel_ids = []
-        self.digital_channels = self.status_channels # Alias the Digitals
-        self.digital_channel_ids = self.status_channel_ids # Alias the Digital Names
+        self.digital_channels = self.status_channels  # Alias the Digitals
+        self.digital_channel_ids = self.status_channel_ids  # Alias Digitals
         self.analog_count = 0
         self.status_count = 0
-        self.digital_count = 0 
+        self.digital_count = 0
         self.frequency = 0.0
 
+        self.year = 1970
+        self.month = 1
+        self.day = 1
+        self.hour = 0
+        self.min = 0
+        self.sec = 0
+        self.msec = 0
+
         self._ignored_channels = []
         self._trig_column = -1
+        self._trig_row = 0
+        self._analog_samp_timedelta = None
+        self._digital_samp_timedelta = None
+        self._properties = {}  # Empty Dictionary of the Keys
 
         # Prepare Data or File if Provided
-        if file is not None:
-            self.load( file=file, encoding=encoding )
-        elif data is not None:
-            self.load_data( data=data, encoding=encoding )
-    
+        if isinstance(file, (str, bytes)):
+            self.load(file=file, encoding=encoding)
+        elif hasattr(file, "read"): # Probably a file-like object?
+            self.load_data(data=file.read(), encoding=encoding)
+        elif isinstance(data, (str, bytes)):
+            self.load_data(data=data, encoding=encoding)
+
     # Define Simple Method to Identify Class Keys
     def _keys(self):
-        """ Capture Class Attributes as Keys """
+        """Capture Class Attributes as Keys."""
         return self.__dict__.keys()
 
     # Define Simple File Extension Validator
     def _validate_extension(self, file):
-        """ Validate Extension is of *.CEV Format """
-        if not os.path.exists( file ):
-            raise ValueError("Argument `file` must be a valid file-path to a CEV file.")
-        filename, ext = os.path.splitext(file)
+        """Validate Extension is of *.CEV Format."""
+        if not os.path.exists(file):
+            raise FileNotFoundError(
+                "Argument `file` must be a valid file-path to a CEV file."
+            )
+        _, ext = os.path.splitext(file)
         if 'CEV' not in ext.upper():
             # Throw Warning to User
             if not self.ignore_warnings:
                 # Capture Pertinent Information
                 parent = inspect.stack()[2]
                 callfile = parent.filename
-                lineno   = parent.lineno
-                function = parent.function
+                lineno = parent.lineno
                 warnings.showwarning(
-                    message=f'File does not appear to use "CEV" extension, instead is "{ext}".',
-                    category=UserWarning,
+                    message=(
+                        'File does not appear to use "CEV" extension,' +
+                        f' instead is "{ext}".'
+                    ),
+                    category=UnexpectedFileExtension,
                     filename=callfile,
                     lineno=lineno,
                 )
         # Proceed Without Exception
-    
+
     # Define Simple Decoder for Data
     def _decode(self, data, encoding):
-        """ Simply Decode the Data Using the Specified Encoding Format """
-        if encoding != None:
+        """Simply Decode the Data Using the Specified Encoding Format."""
+        if encoding is not None:
             return data.decode(encoding, self._decode_opt)
         else:
             return data
-    
+
     # Define Function to Prepare Record and Validate Checksums
     def _prepare_and_validate_record(self):
-        """ Simply Split the Record and Settings, then Evaluate Checksums """
+        """Simply Split the Record and Settings, then Evaluate Checksums."""
         # Split Data
         self.record, self.settings = split_event_and_relay_data(self.data)
-        valid = True # Assume Good
+        invalid_lines = {}
         # Evaluate Record Checksums
-        for line in self.record.split('\n'):
-            if line == '' or line == None or line == ' ':
+        for i, line in enumerate(self.record.split('\n')):
+            if line == '' or (line is None) or line == ' ':
                 continue
             # Collect Line Data and
             content, validity = row_wise_checksum(line)
             self.record_lines.append(content)
-            valid = valid and validity
+            if not validity:
+                invalid_lines[i + 1] = content
         # Throw Warning to User
-        if (not valid) and (not self.ignore_warnings):
+        if invalid_lines and (not self.ignore_warnings):
             # Capture Pertinent Information
             parent = inspect.stack()[2]
             callfile = parent.filename
-            lineno   = parent.lineno
-            function = parent.function
+            lineno = parent.lineno
             warnings.showwarning(
-                message='Record data appears to be malformed, and fails checksum validation.',
-                category=UserWarning,
+                message=(
+                    'Record data appears to be malformed, '
+                    'and fails checksum validation for lines: '
+                    ", ".join([str(ind) for ind in invalid_lines])
+                ),
+                category=MalformedChecksumFailure,
                 filename=callfile,
                 lineno=lineno,
             )
         # Return the Validity Signal
-        return valid
-    
+        return len(invalid_lines) == 0
+
     # Define Internal Test to Identify Header
     def _is_header(self, row_data):
-        """ Simple Test Function to Evaluate Whether Row is Header """
+        """Test Function to Evaluate Whether Row is Header."""
         if row_data.startswith('"') and row_data.endswith('"'):
             return True
         else:
             return False
-    
+
     # Define Internal Test to Identify Data Row
     def _is_data(self, row_data):
-        """ Simple Test Function to Evaluate Whether Row isn't Header """
+        """Test Function to Evaluate Whether Row isn't Header."""
         return not self._is_header(row_data=row_data)
-    
+
     # Define Primary Parsing Function
     def _parse_record(self):
-        """ Primary Parsing Function to Interpret the CEV """
+        """Primary Parsing Function to Interpret the CEV."""
         # Operate on "Row-Pairs" with two Rows at Once to Pair Key with Value
         # Start with Row-Index-Zero (first row), and Assuming Header
         iRow = 0
@@ -491,36 +462,42 @@ class Cev():
         # Manage the Initial Record Data
         while header:
             # Clean and Split the Heading and Content
-            heading_row = self.record_lines[ iRow ].replace('"','').split(',')
-            content_row = self.record_lines[ iRow + 1 ].replace('"','').split(',')
+            heading_row = self.record_lines[iRow].replace('"', '').split(',')
+            content_row = self.record_lines[iRow + 1]
+            content_row = content_row.replace('"', '').split(',')
             if not len(heading_row) == len(content_row):
                 print(heading_row)
                 print(content_row)
-                raise ValueError("CEV may be malformed, heading and data did not match in length.")
+                raise MalformedHeadingDataMismatch(
+                    "CEV may be malformed, heading and data length don't match"
+                )
             # Load the Data into Class Keys
             for key, value in zip(heading_row, content_row):
                 # Verify Attribute and Load
-                key = key.lower()
-                if key in self._keys():
-                    if callable( self.__dict__[key] ):
-                        continue # Don't Overwrite a Callable!
-                # Store the Data
-                self.__dict__[key] = value
-            
+                key_lower = key.lower()
+                if key_lower in self._keys():
+                    if callable(self.__dict__[key_lower]):
+                        continue  # Don't Overwrite a Callable!
+                if (key.find('/') == -1) and (key.find('(') == -1):
+                    # Valid Class Variable Name, Load Directly
+                    self.__dict__[key_lower] = value
+                # Store the Data as a Property
+                self._properties[key] = value
+
             # Check Next Group
             header = (
-                self._is_header( self.record_lines[ iRow + 2 ] ) and
-                self._is_data( self.record_lines[ iRow + 3 ] ) and
-                self._is_header( self.record_lines[ iRow + 4 ] )
+                self._is_header(self.record_lines[iRow + 2]) and
+                self._is_data(self.record_lines[iRow + 3]) and
+                self._is_header(self.record_lines[iRow + 4])
             )
             # Increment Row Index
             iRow += 2
-        
+
         # Following the Primary Header Content, a Single Header Remains
         # with the Analog and Digital Channel Names.
         # Split on either a comma (',') or a space (' ')
-        channels = re.split(r',| ', self.record_lines[ iRow ])
-        is_analog = True # First Channel from Left is Analog
+        channels = re.split(r',| ', self.record_lines[iRow])
+        is_analog = True  # First Channel from Left is Analog
 
         # Identify Channel Names as Analog or Digital
         for i, channel in enumerate(channels):
@@ -529,75 +506,136 @@ class Cev():
                 is_analog = False
                 self._ignored_channels.append(i)
                 self._trig_column = i
-                continue # Don't Track the TRIP Channel
+                continue  # Don't Track the TRIP Channel
+            elif '*' == channel:
+                self._ignored_channels.append(i)
+                continue  # Don't Track Unused Channels
+            elif '' == channel:
+                continue  # Don't Track Empty Channel Names
             # Remove Double Quotes
-            channel = channel.replace('"','')
+            channel = channel.replace('"', '')
             # Channel Must be Valid, Append Name to Either Analog or Digital
             if is_analog:
                 self.analog_channel_ids.append(channel)
             else:
                 self.status_channel_ids.append(channel)
-        
-        iRow += 1 # Increment Past the Data Heading Column
+
+        iRow += 1  # Increment Past the Data Heading Column
 
         # Characterize Number of Channels
         self.analog_count = len(self.analog_channel_ids)
         self.status_count = len(self.status_channel_ids)
 
         # Build the Channel Lists According to Sizes
-        self.analog_channels = [ [] for x in range(self.analog_count) ]
-        self.status_channels = [ [] for x in range(self.status_count) ]
+        self.analog_channels = [[] for x in range(self.analog_count)]
+        self.status_channels = [[] for x in range(self.status_count)]
 
         # Iterate over Data Rows to Load Channels
         numRows = len(self.record_lines)
+        initRow = iRow
         while iRow < numRows:
-            channels = self.record_lines[ iRow ].split(',')
+            channels = self.record_lines[iRow].split(',')
             # Track Analog Quantities
             k = 0
             for i in range(0, self.analog_count):
-                value = float( channels[i] )
-                self.analog_channels[i].append( value )
+                value = float(channels[i])
+                # Verify that Channel Index Shouldn't be Ignored
+                if i not in self._ignored_channels:
+                    self.analog_channels[i].append(value)
                 k = i + 2
-            
+
+            # Identify Trigger Data Row
+            if TRIGGER_KEY_CHAR in channels[k-1]:
+                self._trig_row = iRow - initRow
+
             # Format the Digitals
-            digital_string = channels[k].replace('"','')
-            digitals = hex_bits_from_str( digital_string )
-            
+            digital_string = channels[k].replace('"', '')
+            digitals = []
+            # Iteratively Process the Hex String
+            while len(digital_string) > 0:
+                hex_byte = digital_string[:2]
+                digital_string = digital_string[2:]
+                # Identify the Bits
+                binary_byte = bin(int(hex_byte, base=16))
+                # Use zfill to pad the string with zeros as we want
+                # all 8 digits of the byte.
+                bits_string = binary_byte[2:].zfill(8)
+                digitals.extend([int(bit) for bit in bits_string])
+
             # Track Digital Quantities
             for i in range(0, self.status_count):
-                self.status_channels[i].append( digitals[i] )
-            
-            iRow += 1 # Increment Row Index
-                
-    
+                # Verify that Channel Index (Offset by the TRIG channel)
+                # Shouldn't be Ignored
+                if (i + self._trig_column) not in self._ignored_channels:
+                    self.status_channels[i].append(digitals[i])
+
+            iRow += 1  # Increment Row Index
+
     # Define Event Trigger Time Evaluator
     def _eval_trigger_time(self):
-        """ Simple Function to Use Loaded Time Information to Identify Trigger Time """
+        """Use Time Information to Identify Trigger Time."""
         usec = int(self.msec) * 1000
         self.trigger_time = dt.datetime(
-            year    = int(self.year),
-            month   = int(self.month),
-            day     = int(self.day),
-            hour    = int(self.hour),
-            minute  = int(self.min),
-            second  = int(self.sec),
-            microsecond = usec
+            year=int(self.year),
+            month=int(self.month),
+            day=int(self.day),
+            hour=int(self.hour),
+            minute=int(self.min),
+            second=int(self.sec),
+            microsecond=usec
         )
-    
+
+    # Define Frequency Identifier
+    def _eval_frequency(self):
+        """Identify and load the system nominal frequency."""
+        self.frequency = float(self._properties.get(FREQUENCY_KEY, 60.0))
+
     # Define FID Cleaner
     def _clean_fid(self):
-        """ Simply Store the 'raw' FID in a New Variable, and Clean Existing FID """
-        self.raw_fid = self.fid
-        self.fid = self.fid.split('=')[1]
+        """Store the 'raw' FID in a New Variable, and Clean Existing FID."""
+        try:
+            self.raw_fid = self.fid
+            self.fid = self.fid.split('=')[1]
+        except Exception as err:
+            raise MalformedNoFIDFound(
+                "Failed to load relay FID from CEV"
+            ) from err
+
+    # Define Samples-Per-Cycle Evaluator
+    def _eval_samples_per_cycle(self):
+        """Identify the Samples/Cycle Indicators, Calculate the Deltas."""
+        try:
+            # Extract the Number of Samples per Cycle, Evaluate Milliseconds
+            ms_per_cyc = 1000 / self.frequency
+            analog_ms = ms_per_cyc / float(self._properties[ANALOG_SAMPLES])
+            digital_ms = ms_per_cyc / float(self._properties[DIGITAL_SAMPLES])
+        except KeyError as err:
+            raise MalformedNoSampleNumberFound(
+                "Failed to identify number of samples per cycle"
+            ) from err
+        # Prepare the TimeDeltas
+        self._analog_samp_timedelta = dt.timedelta(milliseconds=analog_ms)
+        self._digital_samp_timedelta = dt.timedelta(milliseconds=digital_ms)
+
+    # Define Timestamp Loader
+    def _eval_timestamps(self):
+        """Evaluate event timestamps."""
+        # Calculate the first timestamp
+        initTimeDelta = self._analog_samp_timedelta * self._trig_row
+        self.time = [self.trigger_time - initTimeDelta]
+        # Iteratively Calculate Remaining Timestamps
+        for _ in range(1, len(self.analog_channels[0])):
+            # Add Timedelta to Most Recent Time Value
+            self.time.append(self.time[-1] + self._analog_samp_timedelta)
 
     # Define File Loader Method
-    def load(self, file, encoding=None):
+    def load(self, file: str, encoding: str = None):
         """
-        *CEV File Loader Method*
-        
+        *CEV File Loader Method*.
+
         Use this method to load a CEV file, and parse its contents into the
         valuable class attributes and structure.
-    
+
         Parameters
         ----------
         file:       str
@@ -609,74 +647,94 @@ class Cev():
                     String specifying the encoding format (if required) in
                     which the file is stored. This may be used for files of
                     format 'utf-8', for example.
-        
+
         See Also
         --------
         load_data       : Load data which has already been read from a file,
                           or is presented as a stream.
-        
+
         Raises
         ------
-        ValueError
+        FileNotFoundError
             If the file cannot be located on the system
-        ValueError
+        MalformedHeadingDataMismatch
             If the file contains a header and content row pair which do not
             share an equal number of columns.
-        
+
         Warns
         -----
-        UserWarning
+        UnexpectedFileExtension
             If the uppercase-cast file extension is not ".CEV"
-        UserWarning
+        MalformedChecksumFailure
             If any of the CEV line-wise checksums do not evaluate successfully
+
+        Examples
+        --------
+        >>> from pycev import CEV
+        >>> # Load a file and parse, directly.
+        >>> record = CEV() # Create the parser instance
+        >>> record.load(file="./event-report.cev")
+        >>> print("Trigger time = {}s".format(record.trigger_time))
         """
-        # Validate file Extension
-        self._validate_extension(file)
-        # Read File with Encoding
-        with open(file, 'r', encoding=encoding) as fObj:
-            self.data = fObj.read() # Gather ALL Data From File
+        # Switch between Handling File-Path, or File-Like-Object
+        if isinstance(file, str):
+            # Validate file Extension
+            self._validate_extension(file)
+            # Read File with Encoding
+            with open(file, 'r', encoding=encoding) as file_obj:
+                self.data = file_obj.read()  # Gather ALL Data From File
+        elif hasattr(file, "read"):
+            self.data = file.read()
         # Process the Data and Load Record
         self.load_data(data=None)
-    
+
     # Define Data Loader Method
-    def load_data(self, data, encoding=None):
+    def load_data(self, data: Union[str, bytes], encoding: str = None):
         """
-        *CEV Data Loader Method*
-        
+        *CEV Data Loader Method*.
+
         Use this method to load the data from a CEV file which has already been
         read, or data which is being streamed to the class (i.e., an active
         connection to a relay). This method will parse the data and load the
         class attributes and structures appropriately.
-    
+
         Parameters
         ----------
-        file:       str
-                    String describing the relative or fully qualified path to
-                    the CEV file that should be read. Optionally used during
-                    class initialization, may also be loaded using the `load`
-                    method.
+        data:       [str, bytes]
+                    String of the full file content.
         encoding:   str, optional
                     String specifying the encoding format (if required) in
                     which the file is stored. This may be used for files of
                     format 'utf-8', for example.
-        
+
         See Also
         --------
         load            : Load data from a CEV file.
-        
+
         Raises
         ------
-        ValueError
+        MalformedHeadingDataMismatch
             If the file contains a header and content row pair which do not
             share an equal number of columns.
-        
+
         Warns
         -----
-        UserWarning
+        MalformedChecksumFailure
             If any of the CEV line-wise checksums do not evaluate successfully
+
+        Examples
+        --------
+        >>> from pycev import CEV
+        >>> # Load a file and parse, directly.
+        >>> record = CEV() # Create the parser instance
+        >>> with open("./event-report.cev", 'r') as file:
+        ...     data = file.read()
+        >>> record.load_data(data=data)
+        >>> print("Trigger time = {}s".format(record.trigger_time))
         """
-        # Method is Called Internally with `data=None`, Don't Try Loading in this Case
-        if data != None:
+        # Method is Called Internally with `data=None`
+        # Don't Try Loading in this Case
+        if data is not None:
             # Decode Data As Needed
             self.data = self._decode(data, encoding=encoding)
         # Prepare Record Information
@@ -685,63 +743,78 @@ class Cev():
         self._parse_record()
         # Determine the Trigger Time
         self._eval_trigger_time()
+        # Identify Nominal System Frequency
+        self._eval_frequency()
         # Clean FID
         self._clean_fid()
-    
+        # Rationalize Number of Samples per Cycle
+        self._eval_samples_per_cycle()
+        # Evaluate the Timestamps
+        self._eval_timestamps()
+
     # Define Method to Access the Analog Channel by Name
-    def get_analog(self, channel_name):
+    def get_analog(self, channel_name: str):
         """
-        *Extract an analog channel by name*
-        
+        *Extract an analog channel by name*.
+
         Use this method to return the list of analog values
         associated with the particular analog channel with
         the specified name.
-        
+
         Parameters
         ----------
         channel_name:   str
                         Name of the analog channel which
                         should be extracted.
-        
+
         Returns
         -------
         channel:    list of float
                     The analog channel values in a zero-based
                     list.
-        
+
         See Also
         --------
         get_status  : Collect the digital channel status for a
                       specified name.
         get_digital : Collect the digital channel status for a
                       specified name.
+
+        Examples
+        --------
+        >>> from pycev import CEV
+        >>> # Load a file and parse, directly.
+        >>> record = CEV() # Create the parser instance
+        >>> record.load(file="./event-report.cev")
+        >>> record.get_analog("FREQ")
+        [...]
         """
         # Identify the Analog Channel Index
         channel_index = self.analog_channel_ids.index(channel_name)
         # Return the Analog Channel
         return self.analog_channels[channel_index]
-    
+
     # Define Method to Access the Digital Channel by Name
-    def get_status(self, channel_name):
+    def get_status(self, channel_name: str):
         """
-        *Extract an digital channel by name*
-        
+        *Extract an digital channel by name*.
+
         Use this method to return the list of digital values
         associated with the particular digital channel with
         the specified name.
-        
+
         Parameters
         ----------
         channel_name:   str
                         Name of the digital channel which
                         should be extracted.
-        
+
         Returns
         -------
-        channel:    list of float
+        channel:    list of bool
                     The digital channel values in a zero-based
                     list.
-        
+
         See Also
         --------
         get_analog  : Collect the analog channel status for a
@@ -750,29 +823,29 @@ class Cev():
                       specified name.
         get_digital : Collect the digital channel status for a
                       specified name.
+
+        Examples
+        --------
+        >>> from pycev import CEV
+        >>> # Load a file and parse, directly.
+        >>> record = CEV() # Create the parser instance
+        >>> record.load(file="./event-report.cev")
+        >>> record.get_status("TRIPLED")
+        [...]
+        >>> record.get_digital("TRIPLED")
+        [...]
         """
         # Identify the Digital Channel Index
         channel_index = self.status_channel_ids.index(channel_name)
         # Return the Digital Channel
         return self.status_channels[channel_index]
-    
+
     # Alias `get_status` to `get_digital`
     get_digital = get_status
 
 
-
 # Alias the Class: `Cev` to `CEV` for Convenience
 CEV = Cev
 
-# Define Simple Builtin Test
-if __name__ == '__main__':
-    print(row_wise_checksum('"SETTINGS","02E1"'))
-    filepath = input("Specify a CEV file to test against: ")
-    x = Cev(file=filepath)
-    print(x.fid)
-    print(x.trigger_time)
-    print(x.analog_channel_ids)
-    print(x.analog_channels[-1])
 
-
-# END
\ No newline at end of file
+# END
